# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for version check done on api methods."""
import functools

from ansys.sherlock.core.errors import SherlockVersionError

_EARLIEST_SUPPORTED_VERSION = 211

# Used in testing. Set this as the version and all version checks will be skipped.
SKIP_VERSION_CHECK = None

# PySherlock release version to the best Sherlock counterpart version
#  0.2.0 : 24R1
#  0.3.0 : 24R1
#  0.4.0 : 24R1
#  0.5.0 : 24R2
#  0.6.0 : 24R2
#  0.7.0 : 25R1
#  0.8.0 : 25R1
#  0.9.0 : 25R2


def require_version(min_version: int = _EARLIEST_SUPPORTED_VERSION, max_version: int = None):
    """Check version of server against expected version."""

    def decorate(func) -> callable:
        """Return wrapped function."""

        @functools.wraps(func)
        # Use functools to keep the doc string associated with the wrapped function, for Sphinx.
        def wrapper(self, *args, **kwargs) -> callable:
            """Wrap outer function."""
            if not hasattr(self, "_server_version"):
                raise SherlockVersionError(
                    "Unable to detect which version of sherlock was launched."
                )

            if self._server_version == SKIP_VERSION_CHECK:
                return func(self, *args, **kwargs)

            cur_version = int(self._server_version)
            nonlocal min_version
            nonlocal max_version

            if max_version is not None and min_version > max_version:
                raise ValueError("min_version is > max_version in version decorator")
            if cur_version < min_version:
                raise SherlockVersionError(
                    "Sherlock version "
                    + str(cur_version)
                    + " too old to use function "
                    + func.__name__
                    + " update server to at least version "
                    + str(min_version)
                )
            if max_version is not None and cur_version > max_version:
                raise SherlockVersionError(
                    "Sherlock version "
                    + str(cur_version)
                    + " is too new. This function "
                    + func.__name__
                    + " deprecated after version "
                    + str(max_version)
                )

            # If no issues have been raised call the wrapped function
            return func(self, *args, **kwargs)

        return wrapper

    return decorate



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Common Service."""

try:
    import SherlockCommonService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2


def basic_str_validator(value: str, field_name: str):
    """Apply basic string validation."""
    if value is None or value == "":
        raise ValueError(field_name + " is invalid because it is None or empty.")
    return value


def optional_str_validator(value: str, field_name: str):
    """Apply optional string validation."""
    if value is None:
        return ""
    return value.strip()


def basic_list_str_validator(value: list, field_name: str):
    """Verify provided list contains at least one item and no empty strings."""
    if not value:
        raise ValueError(f"{field_name} must contain at least one item.")

    for item in value:
        if item is None or item.strip() == "":
            raise ValueError(
                f"{field_name} is invalid because it contains an item that is None or empty."
            )

    return value


class ListUnitsRequestUnitType:
    """Constants for Unit Type in the List Units request."""

    __unit_type = SherlockCommonService_pb2.ListUnitsRequest.UnitType
    ACCEL_DENSITY = __unit_type.ACCEL_DENSITY
    "ACCEL_DENSITY"
    ACCELERATION = __unit_type.ACCELERATION
    "ACCELERATION"
    AREA = __unit_type.AREA
    "AREA"
    BANDWIDTH = __unit_type.BANDWIDTH
    "BANDWIDTH"
    CAPACITANCE = __unit_type.CAPACITANCE
    "CAPACITANCE"
    CTE = __unit_type.CTE
    "CTE"
    CURRENT = __unit_type.CURRENT
    "CURRENT"
    DENSITY = __unit_type.DENSITY
    "DENSITY"
    DISP_DENSITY = __unit_type.DISP_DENSITY
    "DISP_DENSITY"
    FORCE = __unit_type.FORCE
    "FORCE"
    FREQUENCY = __unit_type.FREQUENCY
    "FREQUENCY"
    INDUCTANCE = __unit_type.INDUCTANCE
    "INDUCTANCE"
    LENGTH = __unit_type.LENGTH
    "LENGTH"
    POWER = __unit_type.POWER
    "POWER"
    RESISTANCE = __unit_type.RESISTANCE
    "RESISTANCE"
    SIZE = __unit_type.SIZE
    "SIZE"
    SPECIFIC_HEAT = __unit_type.SPECIFIC_HEAT
    "SPECIFIC_HEAT"
    STRAIN = __unit_type.STRAIN
    "STRAIN"
    STRESS = __unit_type.STRESS
    "STRESS"
    TEMPERATURE = __unit_type.TEMPERATURE
    "TEMPERATURE"
    THERMAL_CONDUCTIVITY = __unit_type.THERMAL_CONDUCTIVITY
    "THERMAL_CONDUCTIVITY"
    THERMAL_RESISTANCE = __unit_type.THERMAL_RESISTANCE
    "THERMAL_RESISTANCE"
    TIME = __unit_type.TIME
    "TIME"
    VELOCITY = __unit_type.VELOCITY
    "VELOCITY"
    VELOCITY_DENSITY = __unit_type.VELOCITY_DENSITY
    "VELOCITY_DENSITY"
    VOLTAGE = __unit_type.VOLTAGE
    "VOLTAGE"
    VOLUME = __unit_type.VOLUME
    "VOLUME"
    WEIGHT = __unit_type.WEIGHT
    "WEIGHT"


class TableDelimiter:
    """Types of delimiters that can be used for exporting tables."""

    __table_delimiter = SherlockCommonService_pb2.TableDelimiter

    COMMA = __table_delimiter.COMMA
    "COMMA"
    SPACE = __table_delimiter.SPACE
    "SPACE"
    TAB = __table_delimiter.TAB
    "TAB"
    SEMICOLON = __table_delimiter.SEMICOLON
    "SEMICOLON"


class Measurement:
    """Contains the properties of the measurement."""

    def __init__(self, value: float, unit: str):
        """Initialize the measurement properties."""
        self.value = value
        """float: Measurement value"""
        self.unit = unit
        """str: Measurement units"""



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for running the gRPC APIs in the Sherlock Common service."""

import grpc

from ansys.sherlock.core.types.common_types import ListUnitsRequestUnitType

try:
    import SherlockCommonService_pb2
    import SherlockCommonService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2_grpc

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import SherlockCommonServiceError, SherlockNoGrpcConnectionException
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.utils.version_check import require_version


class Common(GrpcStub):
    """Contains methods from the Sherlock Common service."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize a gRPC stub for the Sherlock Common service."""
        super().__init__(channel, server_version)
        self.stub = SherlockCommonService_pb2_grpc.SherlockCommonServiceStub(channel)

    @require_version()
    def check(self) -> bool:
        """Perform a health check on the gRPC connection.

        Returns
        -------
        bool
            Whether the Sherlock client is connected via gRPC.
        """
        if not self._is_connection_up():
            LOG.warning("Health check failed.")
            return False
        else:
            LOG.info("Connection is healthy.")
            return True

    @require_version()
    def is_sherlock_client_loading(self) -> bool:
        """Check if the Sherlock client is opened and done initializing.

        Available Since: 2023R2

        Returns
        -------
        bool
            Whether the Sherlock client is opened and done initializing.
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        message = SherlockCommonService_pb2.IsSherlockClientLoadingRequest()
        response = self.stub.isSherlockClientLoading(message)

        if response.value == 0:
            LOG.info("Sherlock client has finished initializing.")
            return True
        else:
            LOG.error("Sherlock client has not finished initializing.")
            return False

    @require_version()
    def exit(self, close_sherlock_client: bool = False):
        """Close the gRPC connection.

        Available Since: 2023R1

        Parameters
        ----------
        close_sherlock_client : bool, optional
            Whether to close the Sherlock client when the gRPC connection is closed. The default
            is ``False``, in which case the Sherlock client remains open when the gRPC connection
            is closed.
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        try:
            exit_message = SherlockCommonService_pb2.ExitRequest()
            exit_message.closeSherlockClient = close_sherlock_client
            response = self.stub.exit(exit_message)
            LOG.info(str(response))
        except SherlockCommonServiceError as err:
            LOG.error("Exit error: ", str(err))

    @require_version()
    def list_units(self, unit_type: ListUnitsRequestUnitType) -> str:
        """List units for a unit type.

        Available Since: 2023R2

        Parameters
        ----------
        unit_type : ListUnitsRequestUnitType
            Unit type.

        Returns
        -------
        str
            Units for the unit type.
        """
        if unit_type == "":
            raise SherlockCommonServiceError(message="Unit type is missing.")

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockCommonService_pb2.ListUnitsRequest(unitType=unit_type)

        try:
            response = self.stub.listUnits(request)
            return_code = response.returnCode
            if return_code.value == -1:
                raise SherlockCommonServiceError(message=return_code.message)
        except SherlockCommonServiceError as e:
            raise e

        return response.units

    @require_version()
    def list_solder_materials(self) -> list[str]:
        """List valid solders.

        Available Since: 2024R1

        Returns
        -------
        list[str]
            Valid solder names.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.common.list_solder_materials()
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockCommonService_pb2.SolderInfoRequest()
        response = self.stub.getSolderInfo(request)

        return response.solders

    @require_version(251)
    def get_sherlock_info(self) -> str:
        """Get server Sherlock version.

        Returns
        -------
        SherlockInfoResponse
            Sherlock information containing
            releaseVersion, defaultProjectDir and isSingleProjectMode flag

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> release_version = sherlock.common.get_sherlock_info().releaseVersion
        >>> default_dir = sherlock.common.get_sherlock_info().defaultProjectDir
        >>> is_single_project = sherlock.common.get_sherlock_info().isSingleProjectMode
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockCommonService_pb2.SherlockInfoRequest()
        response = self.stub.getSherlockInfo(request)

        return response

    @require_version(252)
    def get_solder_info(self) -> SherlockCommonService_pb2.SolderInfoResponse:
        """Get solder data from Sherlock.

        Returns
        -------
        SolderInfoResponse
            All solder information stored in Sherlock for each solder.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> solder_data = sherlock.get_solder_info()
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockCommonService_pb2.SolderInfoRequest()
        return self.stub.getSolderInfo(request)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module for the gRPC connection object."""

import grpc

from ansys.sherlock.core.analysis import Analysis
from ansys.sherlock.core.common import Common
from ansys.sherlock.core.layer import Layer
from ansys.sherlock.core.lifecycle import Lifecycle
from ansys.sherlock.core.model import Model
from ansys.sherlock.core.parts import Parts
from ansys.sherlock.core.project import Project
from ansys.sherlock.core.stackup import Stackup


class Sherlock:
    """Sherlock gRPC connection object."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize Sherlock gRPC connection object."""
        self.common = Common(channel, server_version)
        self.model = Model(channel, server_version)
        self.project = Project(channel, server_version)
        self.lifecycle = Lifecycle(channel, server_version)
        self.layer = Layer(channel, server_version)
        self.stackup = Stackup(channel, server_version)
        self.parts = Parts(channel, server_version)
        self.analysis = Analysis(channel, server_version)



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing all layer management capabilities."""
import grpc

from ansys.sherlock.core.types.layer_types import (
    CircularShape,
    CopyPottingRegionRequest,
    DeletePottingRegionRequest,
    GetICTFixturesPropertiesRequest,
    GetMountPointsPropertiesRequest,
    GetTestPointPropertiesRequest,
    PCBShape,
    PolygonalShape,
    RectangularShape,
    SlotShape,
    UpdateICTFixturesRequest,
    UpdateMountPointsRequest,
    UpdatePottingRegionRequest,
    UpdateTestPointsRequest,
)

try:
    import SherlockCommonService_pb2
    import SherlockLayerService_pb2
    from SherlockLayerService_pb2 import ModelingRegion
    import SherlockLayerService_pb2_grpc
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockCommonService_pb2
    from ansys.api.sherlock.v0 import SherlockLayerService_pb2
    from ansys.api.sherlock.v0 import SherlockLayerService_pb2_grpc
    from ansys.api.sherlock.v0.SherlockLayerService_pb2 import ModelingRegion

from ansys.sherlock.core import LOG
from ansys.sherlock.core.errors import (
    SherlockAddModelingRegionError,
    SherlockAddPottingRegionError,
    SherlockCopyModelingRegionError,
    SherlockDeleteAllICTFixturesError,
    SherlockDeleteAllMountPointsError,
    SherlockDeleteAllTestPointsError,
    SherlockDeleteModelingRegionError,
    SherlockExportAllMountPoints,
    SherlockExportAllTestFixtures,
    SherlockExportAllTestPointsError,
    SherlockExportLayerImageError,
    SherlockListLayersError,
    SherlockNoGrpcConnectionException,
    SherlockUpdateModelingRegionError,
    SherlockUpdateMountPointsByFileError,
    SherlockUpdateTestFixturesByFileError,
    SherlockUpdateTestPointsByFileError,
)
from ansys.sherlock.core.grpc_stub import GrpcStub
from ansys.sherlock.core.utils.version_check import require_version


class Layer(GrpcStub):
    """Module containing all the layer management capabilities."""

    def __init__(self, channel: grpc.Channel, server_version: int):
        """Initialize a gRPC stub for SherlockLayerService."""
        super().__init__(channel, server_version)
        self.stub = SherlockLayerService_pb2_grpc.SherlockLayerServiceStub(channel)

    @require_version(241)
    def add_potting_region(
        self,
        project: str,
        potting_regions: list[
            dict[
                str,
                float
                | str
                | PolygonalShape
                | RectangularShape
                | SlotShape
                | CircularShape
                | PCBShape,
            ]
        ],
    ) -> int:
        """Add one or more potting regions to a given project.

        Available Since: 2024R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        potting_regions: list[dict[str, float | str | PolygonalShape | RectangularShape | SlotShape\
 | CircularShape | PCBShape]]
            Potting region properties consisting of these properties:

            - cca_name: str
                Name of the CCA.
            - potting_id: str
                Potting ID. The default is ``None``.
            - side: str
                The side to add the potting region to. The default is ``None``.
                Options are ``"TOP"``, ``"BOTTOM"``, and ``"BOT"``.
            - material: str
                The potting material. The default is ``None``.
            - potting_units: str
                The potting region units. The default is ``None``.
            - thickness: float
                The potting thickness. The default is ``None``.
            - standoff: float
                The potting standoff. The default is ``None``.
            - shape: PolygonalShape|RectangularShape|SlotShape|CircularShape|PCBShape
                The shape of the potting region.


        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import PolygonalShape
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> polygonal_shape = PolygonalShape(points=[
        >>>     (0, 0),
        >>>     (0, 6.35),
        >>>     (9.77, 0)
        >>> ], rotation=87.8)
        >>> sherlock.layer.add_potting_region(
        >>> "Test",
        >>> [{
        >>>     'cca_name': 'Card',
        >>>     'potting_id': 'Test Region',
        >>>     'side': 'TOP',
        >>>     'material': 'epoxyencapsulant',
        >>>     'potting_units': 'in',
        >>>     'thickness': 0.1,
        >>>     'standoff': 0.2,
        >>>     'shape': polygonal_shape
        >>> },
        >>> ])
        """
        try:
            if project == "":
                raise SherlockAddPottingRegionError(message="Project name is invalid.")

            if not isinstance(potting_regions, list):
                raise SherlockAddPottingRegionError(message="Potting regions argument is invalid.")

            if len(potting_regions) == 0:
                raise SherlockAddPottingRegionError(
                    message="One or more potting regions are required."
                )

            request = SherlockLayerService_pb2.AddPottingRegionRequest(project=project)

            for i, potting_region in enumerate(potting_regions):
                if not isinstance(potting_region, dict):
                    raise SherlockAddPottingRegionError(
                        message=f"Potting region argument is invalid for potting region {i}."
                    )

                if "cca_name" not in potting_region.keys():
                    raise SherlockAddPottingRegionError(
                        message=f"CCA name is missing for potting region {i}."
                    )

                region_request = request.pottingRegions.add()
                region_request.ccaName = potting_region["cca_name"]

                if region_request.ccaName == "":
                    raise SherlockAddPottingRegionError(
                        message=f"CCA name is invalid for potting region {i}."
                    )

                if "potting_id" in potting_region.keys():
                    region_request.pottingID = potting_region["potting_id"]
                if "side" in potting_region.keys():
                    region_request.pottingSide = potting_region["side"]
                if "material" in potting_region.keys():
                    region_request.pottingMaterial = potting_region["material"]
                if "potting_units" in potting_region.keys():
                    region_request.pottingUnits = potting_region["potting_units"]
                if "thickness" in potting_region.keys():
                    region_request.pottingThickness = potting_region["thickness"]
                if "standoff" in potting_region.keys():
                    region_request.pottingStandoff = potting_region["standoff"]
                if "shape" not in potting_region.keys():
                    raise SherlockAddPottingRegionError(
                        message=f"Shape missing for potting region {i}."
                    )

                shape = potting_region["shape"]

                if isinstance(shape, PolygonalShape):
                    if not isinstance(shape.points, list):
                        raise SherlockAddPottingRegionError(
                            message=f"Invalid points argument for potting region {i}."
                        )

                    for j, point in enumerate(shape.points):
                        point_message = region_request.polygonalShape.points.add()

                        if not isinstance(point, tuple) or len(point) != 2:
                            raise SherlockAddPottingRegionError(
                                message=f"Point {j} invalid for potting region {i}."
                            )
                        point_message.x = point[0]
                        point_message.y = point[1]
                    region_request.polygonalShape.rotation = shape.rotation
                elif isinstance(shape, RectangularShape):
                    region_request.rectangularShape.length = shape.length
                    region_request.rectangularShape.width = shape.width
                    region_request.rectangularShape.centerX = shape.center_x
                    region_request.rectangularShape.centerY = shape.center_y
                    region_request.rectangularShape.rotation = shape.rotation
                elif isinstance(shape, SlotShape):
                    region_request.slotShape.length = shape.length
                    region_request.slotShape.width = shape.width
                    region_request.slotShape.nodeCount = shape.node_count
                    region_request.slotShape.centerX = shape.center_x
                    region_request.slotShape.centerY = shape.center_y
                    region_request.slotShape.rotation = shape.rotation
                elif isinstance(shape, CircularShape):
                    region_request.circularShape.diameter = shape.diameter
                    region_request.circularShape.nodeCount = shape.node_count
                    region_request.circularShape.centerX = shape.center_x
                    region_request.circularShape.centerY = shape.center_y
                    region_request.circularShape.rotation = shape.rotation
                elif isinstance(shape, PCBShape):
                    region_request.pCBShape.CopyFrom(SherlockLayerService_pb2.PCBShape())
                else:
                    raise SherlockAddPottingRegionError(
                        message=f"Shape invalid for potting region {i}."
                    )

        except SherlockAddPottingRegionError as e:
            LOG.error(str(e))
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        response = self.stub.addPottingRegion(request)

        try:
            if response.value == -1:
                raise SherlockAddPottingRegionError(response.message)
            else:
                LOG.info(response.message)
                return response.value
        except SherlockAddPottingRegionError as e:
            LOG.error(str(e))
            raise e

    @require_version(251)
    def update_potting_region(
        self, request: UpdatePottingRegionRequest
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        """Update one or more potting regions in a specific project.

        Available Since: 2025R1

        Parameters
        ----------
        request: UpdatePottingRegionRequest
            Contains all the information needed to update one or more potting regions per project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import PolygonalShape
        >>> from ansys.sherlock.core.types.layer_types import PottingRegionUpdateData
        >>> from ansys.sherlock.core.types.layer_types import PottingRegion
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>>
        >>> update1 = PottingRegionUpdateData(
        >>> potting_region_id_to_update=potting_id,
        >>> potting_region=PottingRegionData(
        >>>     cca_name=cca_name,
        >>>     potting_id=potting_id,
        >>>     potting_side=potting_side,
        >>>     potting_material=potting_material,
        >>>     potting_units=potting_units,
        >>>     potting_thickness=potting_thickness,
        >>>     potting_standoff=potting_standoff,
        >>>     shape=PolygonalShape(
        >>>         points=[(0, 1), (5, 1), (5, 5), (1, 5)],
        >>>         rotation=45.0
        >>>     )
        >>> )
        >>> )
        >>> update2 = PottingRegionUpdateData(
        >>> potting_region_id_to_update=potting_id,
        >>> potting_region=PottingRegionData(
        >>>     cca_name=cca_name,
        >>>     potting_id=potting_id,
        >>>     potting_side=potting_side,
        >>>     potting_material=potting_material,
        >>>     potting_units=potting_units,
        >>>     potting_thickness=potting_thickness,
        >>>     potting_standoff=potting_standoff,
        >>>     shape=PolygonalShape(
        >>>         points=[(0, 1), (5, 1), (5, 5), (1, 5)],
        >>>         rotation=0.0
        >>>     )
        >>> )
        >>> )
        >>> example_request = UpdatePottingRegionRequest(
        >>>     "project_name",
        >>>     [
        >>>         update1,
        >>>         update2
        >>>     ]
        >>> )
        >>> return_codes = sherlock.layer.update_potting_region(example_request)
        """
        update_request = request._convert_to_grpc()

        responses = []
        for grpc_return_code in self.stub.updatePottingRegion(update_request):
            responses.append(grpc_return_code)
        return responses

    @require_version(251)
    def copy_potting_region(
        self, request: CopyPottingRegionRequest
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        """Copy one or more potting regions in a specific project.

        Available Since: 2025R1

        Parameters
        ----------
        request: CopyPottingRegionRequest
             Contains all the information needed to copy one or more potting regions per project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import CopyPottingRegionRequest
        >>> from ansys.sherlock.core.types.layer_types import PottingRegionCopyData
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>>
        >>> copy_request_example = CopyPottingRegionRequest(
        >>> project=project,
        >>> potting_region_copy_data=[
        >>>     PottingRegionCopyData(
        >>>         cca_name=cca_name,
        >>>         potting_id=potting_id,
        >>>         copy_potting_id=new_id,
        >>>         center_x=center_x,
        >>>         center_y=center_y
        >>>     ),
        >>>     PottingRegionCopyData(
        >>>         cca_name=cca_name,
        >>>         potting_id=new_id,
        >>>         copy_potting_id=new_id+"1",
        >>>         center_x=center_x,
        >>>         center_y=center_y
        >>>     )
        >>> ]
        >>> )
        >>> responses_example = sherlock.layer.copy_potting_region(copy_request_example)
        """
        copy_request = request._convert_to_grpc()

        responses = []
        for grpc_return_code in self.stub.copyPottingRegion(copy_request):
            responses.append(grpc_return_code)
        return responses

    @require_version(251)
    def delete_potting_region(
        self, request: DeletePottingRegionRequest
    ) -> list[SherlockCommonService_pb2.ReturnCode]:
        """Delete on or more potting regions in a specific project.

        Available Since: 2025R1

        Parameters
        ----------
        request: DeletePottingRegionRequest
             Contains all the information needed to delete one or more potting regions per project.

        Returns
        -------
        list[SherlockCommonService_pb2.ReturnCode]
            Return codes for each request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import DeletePottingRegionRequest
        >>> from ansys.sherlock.core.types.layer_types import PottingRegionDeleteData
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>>
        >>> delete_request_example = DeletePottingRegionRequest(
        >>> project=project,
        >>> potting_region_delete_data=[
        >>>     PottingRegionDeleteData(
        >>>         cca_name=cca_name,
        >>>         potting_id=potting_id
        >>>     )
        >>> ]
        >>> )
        >>> responses_example = sherlock.layer.delete_potting_region(delete_request_example)
        """
        delete_request = request._convert_to_grpc()

        responses = []
        for grpc_return_code in self.stub.deletePottingRegion(delete_request):
            responses.append(grpc_return_code)
        return responses

    @require_version()
    def update_mount_points_by_file(
        self,
        project: str,
        cca_name: str,
        file_path: str,
    ) -> int:
        """Update mount point properties of a CCA from a CSV file.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        file_path: str
            Path for the CSV file with the mount point properties.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_mount_points_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "MountPointImport.csv"
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateMountPointsByFileError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateMountPointsByFileError(message="CCA name is invalid.")
            if file_path == "":
                raise SherlockUpdateMountPointsByFileError(message="File path is required.")
        except SherlockUpdateMountPointsByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        request = SherlockLayerService_pb2.UpdateMountPointsByFileRequest(
            project=project,
            ccaName=cca_name,
            filePath=file_path,
        )

        response = self.stub.updateMountPointsByFile(request)

        return_code = response.returnCode

        try:
            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdateMountPointsByFileError(error_array=response.updateError)

                raise SherlockUpdateMountPointsByFileError(message=return_code.message)
            else:
                LOG.info(return_code.message)
                return return_code.value
        except SherlockUpdateMountPointsByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(222)
    def delete_all_mount_points(self, project: str, cca_name: str) -> int:
        """Delete all mount points for a CCA.

        Available Since: 2022R2

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_mount_points_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "MountPointImport.csv",
        >>> )
        >>> sherlock.layer.delete_all_mount_points("Test", "Card")
        """
        try:
            if project == "":
                raise SherlockDeleteAllMountPointsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockDeleteAllMountPointsError(message="CCA name is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.DeleteAllMountPointsRequest(
                project=project,
                ccaName=cca_name,
            )

            response = self.stub.deleteAllMountPoints(request)

            if response.value == -1:
                raise SherlockDeleteAllMountPointsError(message=response.message)

        except SherlockDeleteAllMountPointsError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(231)
    def delete_all_ict_fixtures(self, project: str, cca_name: str) -> int:
        """Delete all ICT fixtures for a CCA.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_ict_fixtures_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "ICTFixturesImport.csv",
        >>> )
        >>> sherlock.layer.delete_all_ict_fixtures("Test", "Card")
        """
        try:
            if project == "":
                raise SherlockDeleteAllICTFixturesError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockDeleteAllICTFixturesError(message="CCA name is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.DeleteAllICTFixturesRequest(
                project=project,
                ccaName=cca_name,
            )

            response = self.stub.deleteAllICTFixtures(request)

            if response.value == -1:
                raise SherlockDeleteAllICTFixturesError(message=response.message)

        except SherlockDeleteAllICTFixturesError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(231)
    def delete_all_test_points(self, project: str, cca_name: str) -> int:
        """Delete all test points for a CCA.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_test_points_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "TestPointsImport.csv",
        >>> )
        >>> sherlock.layer.delete_all_test_points("Test", "Card")
        """
        try:
            if project == "":
                raise SherlockDeleteAllTestPointsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockDeleteAllTestPointsError(message="CCA name is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.DeleteAllTestPointsRequest(
                project=project,
                ccaName=cca_name,
            )

            response = self.stub.deleteAllTestPoints(request)

            if response.value == -1:
                raise SherlockDeleteAllTestPointsError(message=response.message)

        except SherlockDeleteAllTestPointsError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(231)
    def update_test_points_by_file(
        self,
        project: str,
        cca_name: str,
        file_path: str,
    ) -> int:
        """Update test point properties of a CCA from a CSV file.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        file_path: str
            Path for the CSV file with the test point properties.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_test_points_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "TestPointsImport.csv"
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateTestPointsByFileError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateTestPointsByFileError(message="CCA name is invalid.")
            if file_path == "":
                raise SherlockUpdateTestPointsByFileError(message="File path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.UpdateTestPointsByFileRequest(
                project=project,
                ccaName=cca_name,
                filePath=file_path,
            )

            response = self.stub.updateTestPointsByFile(request)

            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdateTestPointsByFileError(error_array=response.updateError)

                raise SherlockUpdateTestPointsByFileError(message=return_code.message)

            else:
                LOG.info(return_code.message)
                return return_code.value

        except SherlockUpdateTestPointsByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(231)
    def update_test_fixtures_by_file(
        self,
        project: str,
        cca_name: str,
        file_path: str,
    ) -> int:
        """Update test fixture properties of a CCA from a CSV file.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        file_path: str
            Path for the CSV file with the test fixture properties.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.update_test_fixtures_by_file(
        >>>     "Test",
        >>>     "Card",
        >>>     "TestFixturesImport.csv"
        >>> )
        """
        try:
            if project == "":
                raise SherlockUpdateTestFixturesByFileError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockUpdateTestFixturesByFileError(message="CCA name is invalid.")
            if file_path == "":
                raise SherlockUpdateTestFixturesByFileError(message="File path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.UpdateICTFixturesByFileRequest(
                project=project,
                ccaName=cca_name,
                filePath=file_path,
            )

            response = self.stub.updateICTFixturesByFile(request)

            return_code = response.returnCode

            if return_code.value == -1:
                if return_code.message == "":
                    raise SherlockUpdateTestFixturesByFileError(error_array=response.updateError)

                raise SherlockUpdateTestFixturesByFileError(message=return_code.message)

            else:
                LOG.info(return_code.message)
                return return_code.value

        except SherlockUpdateTestFixturesByFileError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(231)
    def export_all_test_points(
        self,
        project: str,
        cca_name: str,
        export_file: str,
        length_units: str = "DEFAULT",
        displacement_units: str = "DEFAULT",
        force_units: str = "DEFAULT",
    ) -> int:
        """Export the test point properties for a CCA.

        Available Since: 2023R1

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        export_file: str
            Full path for the CSV file to export the test points list to.
        length_units: str, optional
            Length units to use when exporting the test points.
            The default is ``DEFAULT``.
        displacement_units: str, optional
            Displacement units to use when exporting the test points.
            The default is ``DEFAULT``.
        force_units: str, optional
            Force units to use when exporting the test points.
            The default is ``DEFAULT``.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.export_all_test_points(
        >>>     "Tutorial Project",
        >>>     "Card",
        >>>     "TestPointsExport.csv",
        >>>     "DEFAULT",
        >>>     "DEFAULT",
        >>>     "DEFAULT"
        >>> )
        """
        try:
            if project == "":
                raise SherlockExportAllTestPointsError(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockExportAllTestPointsError(message="CCA name is invalid.")
            if export_file == "":
                raise SherlockExportAllTestPointsError(message="File path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.ExportAllTestPointsRequest(
                project=project,
                ccaName=cca_name,
                filePath=export_file,
                lengthUnits=length_units,
                displacementUnits=displacement_units,
                forceUnits=force_units,
            )

            response = self.stub.exportAllTestPoints(request)

            if response.value == -1:
                raise SherlockExportAllTestPointsError(message=response.message)

        except SherlockExportAllTestPointsError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(231)
    def export_all_test_fixtures(
        self,
        project: str,
        cca_name: str,
        export_file: str,
        units: str = "DEFAULT",
    ) -> int:
        """Export the test fixture properties for a CCA.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        export_file: str
            Full path for the CSV file to export the text fixtures list to.
        units: str, optional
            Units to use when exporting the test fixtures.
            The default is ``DEFAULT``.


        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.export_all_test_fixtures(
        >>>     "Tutorial Project",
        >>>     "Card",
        >>>     "TestFixturesExport.csv",
        >>>     "DEFAULT"
        >>> )
        """
        try:
            if project == "":
                raise SherlockExportAllTestFixtures(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockExportAllTestFixtures(message="CCA name is invalid.")
            if export_file == "":
                raise SherlockExportAllTestFixtures(message="File path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.ExportAllICTFixturesRequest(
                project=project,
                ccaName=cca_name,
                filePath=export_file,
                units=units,
            )

            response = self.stub.exportAllICTFixtures(request)

            if response.value == -1:
                raise SherlockExportAllTestFixtures(message=response.message)

        except SherlockExportAllTestFixtures as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(222)
    def export_all_mount_points(
        self,
        project: str,
        cca_name: str,
        export_file: str,
        units: str = "DEFAULT",
    ) -> int:
        """Export the mount point properties for a CCA.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.
        export_file: str
            Full path for the CSV file to export the mount points list to.
        units: str, optional
            Units to use when exporting the mount points.
            The default is ``DEFAULT``.


        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> sherlock.layer.export_all_mount_points(
        >>>     "Tutorial Project",
        >>>     "Card",
        >>>     "MountPointsExport.csv",
        >>>     "DEFAULT"
        >>> )
        """
        try:
            if project == "":
                raise SherlockExportAllMountPoints(message="Project name is invalid.")
            if cca_name == "":
                raise SherlockExportAllMountPoints(message="CCA name is invalid.")
            if export_file == "":
                raise SherlockExportAllMountPoints(message="File path is required.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.ExportAllMountPointsRequest(
                project=project,
                ccaName=cca_name,
                filePath=export_file,
                units=units,
            )

            response = self.stub.exportAllMountPoints(request)

            if response.value == -1:
                raise SherlockExportAllMountPoints(message=response.message)

        except SherlockExportAllMountPoints as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(251)
    def add_modeling_region(
        self,
        project: str,
        modeling_regions: list[
            dict[str, bool | float | str | dict[str, bool | float | str] | dict[str, float | str]]
        ],
    ) -> int:
        """
        Add one or more modeling regions to a specific project.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        modeling_regions: list[dict[str, bool | float | str | dict[str, bool | float | str]\
                | dict[str, float | str]]]
            Modeling regions to add. Each dictionary should contain:

            - cca_name: str
                Name of the CCA.
            - region_id: str
                Unique region ID of the modeling region.
            - region_units: str
                Units of the modeling region.
            - model_mode: str
                Mode that specifies how the region is used. Valid values are ``Enabled``,
                ``Disabled`` and ``Excluded``.
            - shape: PolygonalShape|RectangularShape|SlotShape|CircularShape|PCBShape
                The shape of the modeling region.
            - pcb_model_props: dict[str, bool | float | str]
                PCB model parameters consisting of these properties:

                    - export_model_type: str
                        The type of model to be generated for a given modeling region.
                        Valid values are ``Default``, ``Sherlock``, ``Sweep`` and ``None``.
                    - elem_order: str
                        The type of 3D elements to be created for the PCB in the modeling region.
                        Valid values are ``First_Order``, ``Second_Order`` and ``Solid_Shell``.
                    - max_mesh_size: float
                        The maximum size of the mesh to be used in the region.
                    - max_mesh_size_units: str
                        Units for the maximum mesh size.
                    - quads_preferred: bool
                        Whether to generate quad-shaped elements when creating the mesh if true.
            - trace_model_props: dict[str, float | str]
                Trace model parameters consisting of these properties:

                    - trace_model_type : str
                        The specification of whether trace modeling should be performed
                        within the region. Valid values are ``Default``, ``Enabled`` and
                        ``Disabled``.
                    - elem_order: str, optional
                        The type of 3D elements to be created for the PCB in the modeling region.
                        Valid values are ``First_Order``, ``Second_Order`` and ``Solid_Shell``.
                    - trace_mesh_size : float, optional
                        The maximum mesh size to be used in the region when trace modeling
                        is enabled.
                    - trace_mesh_size_units: str, optional
                        Units for the maximum mesh size when trace modeling is enabled.


        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> modeling_regions = [
        >>> {
        >>>     "cca_name": "Card",
        >>>     "region_id": "Region001",
        >>>     "region_units": "mm",
        >>>     "model_mode": "Enabled",
        >>>     "shape": PolygonalShape(points=[
        >>>         (0, 0),
        >>>         (0, 6.35),
        >>>         (9.77, 0)
        >>>     ], rotation=87.8),
        >>>     "pcb_model_props": {
        >>>         "export_model_type": "Sherlock",
        >>>         "elem_order": "First_Order",
        >>>         "max_mesh_size": 0.5,
        >>>         "max_mesh_size_units": "mm",
        >>>         "quads_preferred": True
        >>>     },
        >>>     "trace_model_props": {
        >>>         "trace_model_type": "Enabled",
        >>>         "elem_order": "Second_Order",
        >>>         "trace_mesh_size": 0.3,
        >>>         "trace_mesh_size_units": "mm"
        >>>     }
        >>> }
        >>> ]
        >>> result = sherlock.layer.add_modeling_region("Tutorial Project", modeling_regions)
        """
        try:
            if not project:
                raise SherlockAddModelingRegionError(message="Project name is invalid.")

            if not modeling_regions:
                raise SherlockAddModelingRegionError(message="Modeling regions list is empty.")

            for region in modeling_regions:
                if "cca_name" not in region or not region["cca_name"]:
                    raise SherlockAddModelingRegionError(message="CCA name is invalid.")
                if "region_id" not in region or not region["region_id"]:
                    raise SherlockAddModelingRegionError(message="Region ID is invalid.")
                if "region_units" not in region or not region["region_units"]:
                    raise SherlockAddModelingRegionError(message="Region units are invalid.")
                if "shape" not in region:
                    raise SherlockAddModelingRegionError(message="Shape is missing.")
                elif not isinstance(
                    region["shape"],
                    (
                        PolygonalShape,
                        RectangularShape,
                        SlotShape,
                        CircularShape,
                        PCBShape,
                    ),
                ):
                    raise SherlockAddModelingRegionError(message="Shape is not of a valid type.")

                pcb_model_props = region.get("pcb_model_props", {})
                if pcb_model_props:
                    if (
                        "export_model_type" not in pcb_model_props
                        or pcb_model_props["export_model_type"] == ""
                    ):
                        raise SherlockAddModelingRegionError(
                            message="PCB model export type is invalid."
                        )
                    if "elem_order" not in pcb_model_props or pcb_model_props["elem_order"] == "":
                        raise SherlockAddModelingRegionError(
                            message="PCB element order is invalid."
                        )
                    if "max_mesh_size" not in pcb_model_props or not isinstance(
                        pcb_model_props["max_mesh_size"], float
                    ):
                        raise SherlockAddModelingRegionError(
                            message="PCB max mesh size is invalid."
                        )
                    if "quads_preferred" not in pcb_model_props or not isinstance(
                        pcb_model_props["quads_preferred"], bool
                    ):
                        raise SherlockAddModelingRegionError(
                            message="PCB quads preferred is invalid."
                        )

                trace_model_props = region.get("trace_model_props", {})
                if trace_model_props:
                    if (
                        "trace_model_type" not in trace_model_props
                        or trace_model_props["trace_model_type"] == ""
                    ):
                        raise SherlockAddModelingRegionError(message="Trace model type is invalid.")

                if not self._is_connection_up():
                    raise SherlockNoGrpcConnectionException()

                add_modeling_region_request = SherlockLayerService_pb2.AddModelingRegionRequest()
                add_modeling_region_request.project = project

                for region_request in modeling_regions:
                    modeling_region = add_modeling_region_request.modelingRegions.add()
                    modeling_region.ccaName = region_request["cca_name"]
                    modeling_region.regionId = region_request["region_id"]
                    modeling_region.regionUnits = region_request["region_units"]
                    modeling_region.modelMode = ModelingRegion.ModelingMode.Value(
                        region_request["model_mode"]
                    )

                    shape = region_request["shape"]
                    if isinstance(shape, PolygonalShape):
                        polygonal_shape = modeling_region.polygonalShape
                        for point in shape.points:
                            polygonal_point = polygonal_shape.points.add()
                            polygonal_point.x = point[0]
                            polygonal_point.y = point[1]
                        polygonal_shape.rotation = shape.rotation
                    elif isinstance(shape, RectangularShape):
                        rectangular_shape = modeling_region.rectangularShape
                        rectangular_shape.length = shape.length
                        rectangular_shape.width = shape.width
                        rectangular_shape.centerX = shape.center_x
                        rectangular_shape.centerY = shape.center_y
                        rectangular_shape.rotation = shape.rotation
                    elif isinstance(shape, SlotShape):
                        slot_shape = modeling_region.slotShape
                        slot_shape.length = shape.length
                        slot_shape.width = shape.width
                        slot_shape.nodeCount = shape.node_count
                        slot_shape.centerX = shape.center_x
                        slot_shape.centerY = shape.center_y
                        slot_shape.rotation = shape.rotation
                    elif isinstance(shape, CircularShape):
                        circular_shape = modeling_region.circularShape
                        circular_shape.diameter = shape.diameter
                        circular_shape.nodeCount = shape.node_count
                        circular_shape.centerX = shape.center_x
                        circular_shape.centerY = shape.center_y
                        circular_shape.rotation = shape.rotation
                    else:
                        raise SherlockAddModelingRegionError(
                            message="Shape is not of a valid type."
                        )

                    export_model_type = ModelingRegion.PCBModelingProperties.ExportModelType
                    pcb_model_props = region_request.get("pcb_model_props", {})
                    modeling_region.pcbModelProps.exportModelType = getattr(
                        export_model_type,
                        pcb_model_props["export_model_type"],
                    )
                    modeling_region.pcbModelProps.elemOrder = getattr(
                        ModelingRegion.ElementOrder,
                        pcb_model_props["elem_order"],
                    )
                    modeling_region.pcbModelProps.maxMeshSize = pcb_model_props["max_mesh_size"]
                    modeling_region.pcbModelProps.maxMeshSizeUnits = pcb_model_props[
                        "max_mesh_size_units"
                    ]
                    modeling_region.pcbModelProps.quadsPreferred = pcb_model_props[
                        "quads_preferred"
                    ]

                    trace_modeling_type = ModelingRegion.TraceModelingProperties.TraceModelingType
                    trace_model_props = region_request.get("trace_model_props", {})
                    modeling_region.traceModelProps.traceModelType = getattr(
                        trace_modeling_type,
                        trace_model_props["trace_model_type"],
                    )
                    if "elem_order" in trace_model_props:
                        modeling_region.traceModelProps.elemOrder = getattr(
                            ModelingRegion.ElementOrder,
                            trace_model_props["elem_order"],
                        )
                    if "trace_mesh_size" in trace_model_props:
                        modeling_region.traceModelProps.traceMeshSize = trace_model_props[
                            "trace_mesh_size"
                        ]
                    if "trace_mesh_size_units" in trace_model_props:
                        modeling_region.traceModelProps.traceMeshSizeUnits = trace_model_props[
                            "trace_mesh_size_units"
                        ]

                return_code = self.stub.addModelingRegion(add_modeling_region_request)
                if return_code.value != 0:
                    raise SherlockAddModelingRegionError(message=return_code.message)

                return return_code.value

        except SherlockAddModelingRegionError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(251)
    def update_modeling_region(
        self,
        project: str,
        modeling_regions: list[
            dict[
                str,
                bool | float | str | dict[str, bool | float | str] | dict[str, bool | float | str],
            ]
        ],
    ) -> int:
        """
        Update one or more modeling regions in a specific project.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        modeling_regions : list[dict]
            Modeling regions to update. Each dictionary should contain:

            - cca_name: str
                Name of the CCA.
            - region_id: str
                Unique region ID of the modeling region.
            - region_units: str
                Units of the modeling region.
            - model_mode: str
                Mode that specifies how the region is used. Valid values are ``Enabled``,
                ``Disabled`` and ``Excluded``.
            - shape: PolygonalShape|RectangularShape|SlotShape|CircularShape|PCBShape
                The shape of the modeling region.
            - pcb_model_props: dict[str, bool | float | str]
                PCB model parameters consisting of these properties:

                    - export_model_type: str
                        The type of model to be generated for a given modeling region.
                        Valid values are ``Default``, ``Sherlock``, ``Sweep`` and ``None``.
                    - elem_order: str
                        The type of 3D elements to be created for the PCB in the modeling region.
                        Valid values are ``First_Order``, ``Second_Order`` and ``Solid_Shell``.
                    - max_mesh_size: float
                        The maximum size of the mesh to be used in the region.
                    - max_mesh_size_units: str
                        Units for the maximum mesh size.
                    - quads_preferred: bool
                        Whether to generate quad-shaped elements when creating the mesh if true.
            - trace_model_props: dict[str, float | str]
                Trace model parameters consisting of these properties:

                    - trace_model_type: str
                        The specification of whether trace modeling should be performed
                        within the region. Valid values are ``Default``, ``Enabled`` and
                        ``Disabled``.
                    - elem_order: str, optional
                        The type of 3D elements to be created for the PCB in the modeling region.
                        Valid values are ``First_Order``, ``Second_Order`` and ``Solid_Shell``.
                    - trace_mesh_size: float, optional
                        The maximum mesh size to be used in the region when trace modeling
                        is enabled.
                    - trace_mesh_size_units: str, optional
                        Units for the maximum mesh size when trace modeling is enabled.
            - region_id_replacement: str, optional
                Represents a unique region id that will replace the existing regionId value during
                a modeling region update if a value exists.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> modeling_regions = [
        >>>     {
        >>>         "cca_name": "Card",
        >>>         "region_id": "Region001",
        >>>         "region_units": "mm",
        >>>         "model_mode": "Enabled",
        >>>         "shape": PolygonalShape(points=[(0, 0), (1, 1)], rotation=0),
        >>>         "pcb_model_props": {
        >>>             "export_model_type": "Sherlock",
        >>>             "elem_order": "Second_Order",
        >>>             "max_mesh_size": 0.5,
        >>>             "max_mesh_size_units": "mm",
        >>>             "quads_preferred": True,
        >>>         },
        >>>         "trace_model_props": {
        >>>             "trace_model_type": "Enabled",
        >>>             "elem_order": "Second_Order",
        >>>             "trace_mesh_size": 0.1,
        >>>             "trace_mesh_size_units": "mm",
        >>>         },
        >>>         "region_id_replacement": "NewRegion001",
        >>>     }
        >>> ]
        >>> result = sherlock.layer.update_modeling_region("Tutorial Project", modeling_regions)
        """
        try:
            if not project:
                raise SherlockUpdateModelingRegionError(message="Project name is invalid.")

            if not modeling_regions:
                raise SherlockUpdateModelingRegionError(message="Modeling regions list is empty.")

            for region in modeling_regions:
                if "cca_name" not in region:
                    raise SherlockUpdateModelingRegionError(message="CCA name is invalid.")
                if "region_id" not in region:
                    raise SherlockUpdateModelingRegionError(message="Region ID is invalid.")
                if "region_units" not in region:
                    raise SherlockUpdateModelingRegionError(message="Region units are invalid.")
                if "shape" not in region:
                    raise SherlockUpdateModelingRegionError(message="Shape is missing.")
                elif not isinstance(
                    region["shape"],
                    (
                        PolygonalShape,
                        RectangularShape,
                        SlotShape,
                        CircularShape,
                        PCBShape,
                    ),
                ):
                    raise SherlockUpdateModelingRegionError(message="Shape is not of a valid type.")

                pcb_model_props = region.get("pcb_model_props", {})
                if pcb_model_props:
                    if (
                        "export_model_type" not in pcb_model_props
                        or pcb_model_props["export_model_type"] == ""
                    ):
                        raise SherlockUpdateModelingRegionError(
                            message="PCB model export type is invalid."
                        )
                    if "elem_order" not in pcb_model_props or pcb_model_props["elem_order"] == "":
                        raise SherlockUpdateModelingRegionError(
                            message="PCB element order is invalid."
                        )
                    if "max_mesh_size" not in pcb_model_props or not isinstance(
                        pcb_model_props["max_mesh_size"], float
                    ):
                        raise SherlockUpdateModelingRegionError(
                            message="PCB max mesh size is invalid."
                        )
                    if "quads_preferred" not in pcb_model_props or not isinstance(
                        pcb_model_props["quads_preferred"], bool
                    ):
                        raise SherlockUpdateModelingRegionError(
                            message="PCB quads preferred is invalid."
                        )

                trace_model_props = region.get("trace_model_props", {})
                if trace_model_props:
                    if (
                        "trace_model_type" not in trace_model_props
                        or trace_model_props["trace_model_type"] == ""
                    ):
                        raise SherlockUpdateModelingRegionError(
                            message="Trace model type is invalid."
                        )

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            update_modeling_region_request = SherlockLayerService_pb2.UpdateModelingRegionRequest()
            update_modeling_region_request.project = project

            for region_request in modeling_regions:
                modeling_region = update_modeling_region_request.modelingRegions.add()
                modeling_region.ccaName = region_request["cca_name"]
                modeling_region.regionId = region_request["region_id"]
                modeling_region.regionUnits = region_request["region_units"]
                modeling_region.modelMode = ModelingRegion.ModelingMode.Value(
                    region_request["model_mode"]
                )

                shape = region_request["shape"]
                if isinstance(shape, PolygonalShape):
                    polygonal_shape = modeling_region.polygonalShape
                    for point in shape.points:
                        polygonal_point = polygonal_shape.points.add()
                        polygonal_point.x = point[0]
                        polygonal_point.y = point[1]
                    polygonal_shape.rotation = shape.rotation
                elif isinstance(shape, RectangularShape):
                    rectangular_shape = modeling_region.rectangularShape
                    rectangular_shape.length = shape.length
                    rectangular_shape.width = shape.width
                    rectangular_shape.centerX = shape.center_x
                    rectangular_shape.centerY = shape.center_y
                    rectangular_shape.rotation = shape.rotation
                elif isinstance(shape, SlotShape):
                    slot_shape = modeling_region.slotShape
                    slot_shape.length = shape.length
                    slot_shape.width = shape.width
                    slot_shape.nodeCount = shape.node_count
                    slot_shape.centerX = shape.center_x
                    slot_shape.centerY = shape.center_y
                    slot_shape.rotation = shape.rotation
                elif isinstance(shape, CircularShape):
                    circular_shape = modeling_region.circularShape
                    circular_shape.diameter = shape.diameter
                    circular_shape.nodeCount = shape.node_count
                    circular_shape.centerX = shape.center_x
                    circular_shape.centerY = shape.center_y
                    circular_shape.rotation = shape.rotation

                export_model_type = ModelingRegion.PCBModelingProperties.ExportModelType
                pcb_model_props = region_request.get("pcb_model_props", {})
                modeling_region.pcbModelProps.exportModelType = getattr(
                    export_model_type,
                    pcb_model_props["export_model_type"],
                )
                modeling_region.pcbModelProps.elemOrder = getattr(
                    ModelingRegion.ElementOrder,
                    pcb_model_props["elem_order"],
                )
                modeling_region.pcbModelProps.maxMeshSize = pcb_model_props["max_mesh_size"]
                modeling_region.pcbModelProps.maxMeshSizeUnits = pcb_model_props[
                    "max_mesh_size_units"
                ]
                modeling_region.pcbModelProps.quadsPreferred = pcb_model_props["quads_preferred"]

                trace_modeling_type = ModelingRegion.TraceModelingProperties.TraceModelingType
                trace_model_props = region_request.get("trace_model_props", {})
                modeling_region.traceModelProps.traceModelType = getattr(
                    trace_modeling_type,
                    trace_model_props["trace_model_type"],
                )
                if "elem_order" in trace_model_props:
                    modeling_region.traceModelProps.elemOrder = getattr(
                        ModelingRegion.ElementOrder,
                        trace_model_props["elem_order"],
                    )
                if "trace_mesh_size" in trace_model_props:
                    modeling_region.traceModelProps.traceMeshSize = trace_model_props[
                        "trace_mesh_size"
                    ]
                if "trace_mesh_size_units" in trace_model_props:
                    modeling_region.traceModelProps.traceMeshSizeUnits = trace_model_props[
                        "trace_mesh_size_units"
                    ]

                if "region_id_replacement" in region_request:
                    modeling_region.regionIdReplacement = region_request["region_id_replacement"]

            return_code = self.stub.updateModelingRegion(update_modeling_region_request)
            if return_code.value != 0:
                raise SherlockUpdateModelingRegionError(message=return_code.message)

            return return_code.value

        except SherlockUpdateModelingRegionError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(251)
    def copy_modeling_region(
        self,
        project: str,
        copy_regions: list[dict[str, float | str]],
    ) -> int:
        """
        Copy one or more modeling regions in a specific project.

        Parameters
        ----------
        project : str
            Name of the Sherlock project.
        copy_regions : list[dict[str, float | str]]
            Modeling regions to copy along with their corresponding "copy to" parameters.
            Each dictionary should contain:

            - cca_name : str
                Name of the CCA.
            - region_id : str
                Region ID of the existing modeling region to copy.
            - region_id_copy : str
                Region ID of the modeling region copy. Must be unique.
            - center_x : float
                The center x coordinate of the modeling region copy. Used for location placement in
                the Layer Viewer.
            - center_y : float
                The center y coordinate of the modeling region copy. Used for location placement in
                the Layer Viewer.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card",
        >>> )
        >>> modeling_regions = [
        >>>     {
        >>>         "cca_name": "Card",
        >>>         "region_id": "Region001",
        >>>         "region_id_copy": "RegionCopy001",
        >>>         "center_x": 10.0,
        >>>         "center_y": 20.0,
        >>>     }
        >>> ]
        >>> result = sherlock.layer.copy_modeling_region("Tutorial Project", modeling_regions)
        """
        try:
            if not project:
                raise SherlockCopyModelingRegionError(message="Project name is invalid.")

            if not copy_regions:
                raise SherlockCopyModelingRegionError(message="Copy regions list is empty.")

            for region in copy_regions:
                if "cca_name" not in region or region["cca_name"] == "":
                    raise SherlockCopyModelingRegionError(message="CCA name is invalid.")
                if "region_id" not in region or region["region_id"] == "":
                    raise SherlockCopyModelingRegionError(message="Region ID is invalid.")
                if "region_id_copy" not in region or region["region_id_copy"] == "":
                    raise SherlockCopyModelingRegionError(message="Region ID copy is invalid.")
                if "center_x" not in region or not isinstance(region["center_x"], float):
                    raise SherlockCopyModelingRegionError(message="Center X coordinate is invalid.")
                if "center_y" not in region or not isinstance(region["center_y"], float):
                    raise SherlockCopyModelingRegionError(message="Center Y coordinate is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            copy_regions_request = []

            for region in copy_regions:
                copy_region = (
                    SherlockLayerService_pb2.CopyModelingRegionRequest.CopyModelingRegionInfo(
                        ccaName=region["cca_name"],
                        regionId=region["region_id"],
                        regionIdCopy=region["region_id_copy"],
                        centerX=region["center_x"],
                        centerY=region["center_y"],
                    )
                )
                copy_regions_request.append(copy_region)

            request = SherlockLayerService_pb2.CopyModelingRegionRequest(
                project=project,
                copyRegions=copy_regions_request,
            )

            response = self.stub.copyModelingRegion(request)

            if response.value == -1:
                raise SherlockCopyModelingRegionError(message=response.message)

            return response.value

        except SherlockCopyModelingRegionError as e:
            for error in e.str_itr():
                LOG.error(error)
            raise e

    @require_version(251)
    def delete_modeling_region(self, project: str, delete_regions: list[dict[str, str]]) -> int:
        """Delete one or more modeling regions for a specific project.

        Parameters
        ----------
        project: str
            Name of the Sherlock project.
        delete_regions: list[dict[str, str]]
            Modeling regions to delete. Each dictionary should contain:
            - "cca_name": str, Name of the CCA.
            - "region_id": str, Unique region ID of the modeling region to delete.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Test",
        >>>     cca_name="Card",
        >>> )
        >>> modeling_regions = [{"cca_name": "Card", "region_id": "12345"}]
        >>> sherlock.layer.delete_modeling_region("Test", modeling_regions)
        """
        try:
            if project == "":
                raise SherlockDeleteModelingRegionError(message="Project name is invalid.")
            if not isinstance(delete_regions, list):
                raise SherlockDeleteModelingRegionError(message="Delete regions should be a list.")
            if not delete_regions:
                raise SherlockDeleteModelingRegionError(message="Delete regions list is empty.")

            for region in delete_regions:
                if not isinstance(region, dict):
                    raise SherlockDeleteModelingRegionError(
                        message="Each region should be a dictionary."
                    )
                if "cca_name" not in region or region["cca_name"] == "":
                    raise SherlockDeleteModelingRegionError(message="CCA name is invalid.")
                if "region_id" not in region or region["region_id"] == "":
                    raise SherlockDeleteModelingRegionError(message="Region ID is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            delete_regions_info = [
                SherlockLayerService_pb2.DeleteModelingRegionRequest.DeleteModelingRegionInfo(
                    ccaName=region["cca_name"], regionId=region["region_id"]
                )
                for region in delete_regions
            ]

            request = SherlockLayerService_pb2.DeleteModelingRegionRequest(
                project=project, deleteRegions=delete_regions_info
            )

            response = self.stub.deleteModelingRegion(request)

            if response.value == -1:
                raise SherlockDeleteModelingRegionError(message=response.message)

        except SherlockDeleteModelingRegionError as e:
            LOG.error(str(e))
            raise e

        return response.value

    @require_version(252)
    def list_layers(self, project, cca_name) -> list:
        """List all layers as seen in the Layer Viewer for a specific project CCA.

        Parameters
        ----------
        project : str
            Name of the Sherlock project.
        cca_name: str
            Name of the CCA.

        Returns
        -------
        layer_infos: list[dict[str, list]]
            The layers as seen in the Layer Viewer for the given project CCA.
            Each dictionary should contain:

            - "layer_folder": str, name of layer_folder enum.
            - "layers": list, list of names of layers under this folder

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card"
        >>> )
        >>> sherlock.layer.list_layers(
        >>>     project="Tutorial Project",
        >>>     cca_name="Card"
        >>> )
        """
        try:
            if project == "":
                raise SherlockListLayersError(message="Project name is invalid.")

            if cca_name == "":
                raise SherlockListLayersError(message="CCA name is invalid.")

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            request = SherlockLayerService_pb2.ListLayersRequest(project=project, ccaName=cca_name)

            response = self.stub.listLayers(request)
            if response.returnCode.value == -1:
                raise SherlockListLayersError(response.returnCode.message)

            return response.layerInfos

        except SherlockListLayersError as e:
            LOG.error(str(e))
            raise e

    @require_version(252)
    def export_layer_image(
        self, project: str, cca_name: str, export_layers: list[dict[str, bool | int | str | list]]
    ) -> int:
        r"""
        Export one or more 2D Layer Viewer images from a project CCA.

        Parameters
        ----------
        project : str
            Name of the Sherlock project.
        cca_name : str
            Name of the CCA.
        export_layers : export_layer_image_info
            List of parameters for the export image specified.
            Each dictionary should contain:

            - components_enabled: bool, optional
                Displays the components in the export image. Default to true when not provided.
            - labels_enabled: bool, optional
                Displays the component reference designators in the export image. Default to true
                when not provided.
            - leads_enabled : bool, optional
                Displays the component leads and solder balls in the export image. Default to true
                when not provided.
            - axes_enabled : bool
                Displays the x and y axes in the export image.
            - grid_enabled : bool
                Displays a grid in the export image.
            - layer_infos: list[dict[enum, list]]
                The layers as seen in the Layer Viewer for the given project CCA.
                Each dictionary should contain:

                - "layer_folder": enum, layer_folder enum.
                - "layers": list, list of names of layers under this folder
            - file_path : str
                Full file path of the export image.
            - image_height : int
                The export image height.
            - image_width : int
                The export image width.
            - overwrite_existing_file : bool
                If the file path already exists, overwrite the file if this is set to true.

        Returns
        -------
        int
            Status code of the response. 0 for success.

        Examples
        --------
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> sherlock.project.import_odb_archive(
        >>>     "ODB++ Tutorial.tgz",
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     True,
        >>>     project="Tutorial Project",
        >>>     cca_name="Card"
        >>> )
        >>> layer_infos = [
        >>> { "layer_folder": "Components",
        >>>   "layers": ["comp-top"]},
        >>> { "layer_folder": "Harmonic_Vibe",
        >>>   "layers":["HV Disp @ 203.39 Hz"]}
        >>> ]
        >>> export_layers = [
        >>> {
        >>>     "components_enabled": True,
        >>>     "labels_enabled": True,
        >>>     "leads_enabled": True,
        >>>     "axes_enabled": True,
        >>>     "grid_enabled": True,
        >>>     "layer_infos": layer_infos,
        >>>     "file_path": "C:\\Users\\user_id\\Downloads\\SH-image.jpg",
        >>>     "image_height": 600,
        >>>     "image_width": 800,
        >>>     "overwrite_existing_file": True
        >>> }
        >>> ]
        >>> sherlock.layer.export_layer_image("Tutorial Project", "Card", export_layers)
        """
        try:
            if project == "":
                raise SherlockExportLayerImageError(message="Project name is invalid.")

            if cca_name == "":
                raise SherlockExportLayerImageError(message="CCA name is invalid.")

            for export_layer in export_layers:
                if "axes_enabled" not in export_layer:
                    raise SherlockExportLayerImageError(message="Axes Enabled is invalid.")
                if "grid_enabled" not in export_layer:
                    raise SherlockExportLayerImageError(message="Grid Enabled is invalid.")
                if "layer_infos" not in export_layer or len(export_layer["layer_infos"]) == 0:
                    raise SherlockExportLayerImageError(message="Layer info is invalid.")
                if "file_path" not in export_layer or export_layer["file_path"] == "":
                    raise SherlockExportLayerImageError(message="File Path is invalid.")
                if "image_height" not in export_layer or not isinstance(
                    export_layer["image_height"], int
                ):
                    raise SherlockExportLayerImageError(message="Image Height is invalid.")
                if "image_width" not in export_layer or not isinstance(
                    export_layer["image_width"], int
                ):
                    raise SherlockExportLayerImageError(message="Image Width is invalid.")
                if "overwrite_existing_file" not in export_layer:
                    raise SherlockExportLayerImageError(
                        message="Overwrite Existing File is " "invalid."
                    )

            if not self._is_connection_up():
                raise SherlockNoGrpcConnectionException()

            export_layer_image_request = []

            for export_layer in export_layers:
                if "components_enabled" in export_layer:
                    components_enabled_val = export_layer["components_enabled"]
                else:
                    components_enabled_val = None

                if "labels_enabled" in export_layer:
                    labels_enabled_val = export_layer["labels_enabled"]
                else:
                    labels_enabled_val = None

                if "leads_enabled" in export_layer:
                    leads_enabled_val = export_layer["leads_enabled"]
                else:
                    leads_enabled_val = None

                layer_info_request = []
                for layer in export_layer["layer_infos"]:
                    layer_info = SherlockLayerService_pb2.LayerInfo(
                        layerFolder=layer["layer_folder"], layers=layer["layers"]
                    )
                    layer_info_request.append(layer_info)

                export_layer_info = (
                    SherlockLayerService_pb2.ExportLayerImageRequest.ExportLayerImageInfo(
                        componentsEnabled=components_enabled_val,
                        labelsEnabled=labels_enabled_val,
                        leadsEnabled=leads_enabled_val,
                        axesEnabled=export_layer["axes_enabled"],
                        gridEnabled=export_layer["grid_enabled"],
                        layerInfos=layer_info_request,
                        filePath=export_layer["file_path"],
                        imageHeight=export_layer["image_height"],
                        imageWidth=export_layer["image_width"],
                        overwriteExistingFile=export_layer["overwrite_existing_file"],
                    )
                )
                export_layer_image_request.append(export_layer_info)

            request = SherlockLayerService_pb2.ExportLayerImageRequest(
                project=project, ccaName=cca_name, exportLayers=export_layer_image_request
            )

            responses = []
            for response in self.stub.exportLayerImage(request):
                responses.append(response)
            return responses

        except SherlockExportLayerImageError as e:
            LOG.error(str(e))
            raise e

    @require_version(261)
    def get_test_point_props(
        self, request: GetTestPointPropertiesRequest
    ) -> list[SherlockLayerService_pb2.GetTestPointPropertiesResponse]:
        """Return the properties for each test point given a comma-separated list of test point IDs.

        Available Since: 2026R1

        Parameters
        ----------
        request: GetTestPointPropertiesRequest
             Contains all the information needed to return the properties for one or more test
             points.

        Returns
        -------
        list[SherlockCommonService_pb2.GetTestPointPropertiesResponse]
            Properties for each test point that correspond to the reference designators.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import GetTestPointPropertiesRequest
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> request = layer_types.GetTestPointPropertiesRequest(
        >>>    project = "Test Point Test Project"
        >>>    cca_name = "Main Board"
        >>>    test_point_ids = "TP1,TP2"
        >>> )
        >>> responses = layer.get_test_point_props(request)
        """
        get_test_point_props_request = request._convert_to_grpc()
        responses = []
        for get_test_point_props_response in self.stub.getTestPointProperties(
            get_test_point_props_request
        ):
            responses.append(get_test_point_props_response)
        return responses

    @require_version(261)
    def get_ict_fixtures_props(
        self, request: GetICTFixturesPropertiesRequest
    ) -> SherlockLayerService_pb2.GetICTFixturesPropertiesResponse:
        """Return the properties for each ICT fixture given a comma-separated list of fixture IDs.

        Available Since: 2026R1

        Parameters
        ----------
        request: GetICTFixturesPropertiesRequest
             Contains all the information needed to return the properties for one or more ICT
             fixtures.

        Returns
        -------
        SherlockCommonService_pb2.GetICTFixturesPropertiesResponse
            Properties for each ICT fixture that correspond to the reference designators.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import GetICTFixturesPropertiesRequest
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> request = layer_types.GetICTFixturesPropertiesRequest(
        >>>    project = "Tutorial Project"
        >>>    cca_name = "Main Board"
        >>>    ict_fixtures_ids = "F1,F2"
        >>> )
        >>> response = layer.get_ict_fixtures_props(request)
        """
        get_ict_fixture_props_request = request._convert_to_grpc()
        response = self.stub.getICTFixturesProperties(get_ict_fixture_props_request)

        return response

    @require_version(261)
    def update_test_points(
        self, request: UpdateTestPointsRequest
    ) -> SherlockLayerService_pb2.UpdateTestPointsResponse:
        """Update test point properties of a CCA from input parameters.

        Available Since: 2026R1

        Parameters
        ----------
        request: UpdateTestPointsRequest
            Contains all the information needed to update the properties for one or more
            test points.

        Returns
        -------
        SherlockCommonService_pb2.UpdateTestPointsResponse
            A status code and message for the update test points request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import UpdateTestPointsRequest,
        >>> TestPointProperties
        >>> from ansys.api.sherlock.v0 import SherlockLayerService_pb2
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> test_point = TestPointProperties(
        >>>     id="TP1",
        >>>     side="BOTTOM",
        >>>     units="in",
        >>>     center_x=1.0,
        >>>     center_y=0.5,
        >>>     radius=0.2,
        >>>     load_type=SherlockLayerService_pb2.TestPointProperties.LoadType.Force,
        >>>     load_value=3.0,
        >>>     load_units="ozf",
        >>> )
        >>> response = sherlock.layer.update_test_points(UpdateTestPointsRequest(
        >>> project="Tutorial Project",
        >>> cca_name="Main Board",
        >>> update_test_points=[test_point],
        >>> ))
        """
        update_request = request._convert_to_grpc()
        response = self.stub.updateTestPoints(update_request)
        return response

    @require_version(261)
    def update_ict_fixtures(
        self, request: UpdateICTFixturesRequest
    ) -> SherlockLayerService_pb2.UpdateICTFixturesResponse:
        """Update ict fixture properties of a CCA from input parameters.

        Available Since: 2026R1

        Parameters
        ----------
        request: UpdateICTFixturesRequest
            Contains all the information needed to update the properties for one or more
            ict fixtures.

        Returns
        -------
        SherlockCommonService_pb2.UpdateICTFixturesResponse
            A status code and message for the update ict fixtures request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import UpdateICTFixturesRequest,
        >>> ICTFixtureProperties
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> fixture = ICTFixtureProperties(
        >>>     id="F1",
        >>>     type="Mount Hole",
        >>>     units="in",
        >>>     side="TOP",
        >>>     height="0.0",
        >>>     material="GOLD",
        >>>     state="DISABLED",
        >>>     shape="Slot",
        >>>     x="0.3",
        >>>     y="-0.4",
        >>>     length="1.0",
        >>>     width="0.2",
        >>>     diameter="0.0",
        >>>     nodes="10",
        >>>     rotation="15",
        >>>     polygon="",
        >>>     boundary="Outline",
        >>>     constraints="X-axis translation|Z-axis translation",
        >>>     chassis_material="SILVER",
        >>> )
        >>> response = sherlock.layer.update_ict_fixtures(UpdateICTFixturesRequest(
        >>> project="Tutorial Project",
        >>> cca_name="Main Board",
        >>> update_fixtures=[fixture],
        >>> ))
        """
        update_request = request._convert_to_grpc()
        response = self.stub.updateICTFixtures(update_request)
        return response

    @require_version(261)
    def get_mount_point_props(
        self, request: GetMountPointsPropertiesRequest
    ) -> SherlockLayerService_pb2.GetMountPointsPropertiesResponse:
        """
        Return the properties for each mount point given a comma-separated list of mount point IDs.

        Available Since: 2026R1

        Parameters
        ----------
        request: GetmountPointsPropertiesRequest
             Contains all the information needed to return the properties for one or more mount
             points.

        Returns
        -------
        SherlockCommonService_pb2.GetMountPointsPropertiesResponse
            Properties for each mount point that correspond to the reference designators.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import GetMountPointsPropertiesRequest
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> request = layer_types.GetMountPointsPropertiesRequest(
        >>>    project = "Tutorial Project"
        >>>    cca_name = "Main Board"
        >>>    mount_point_ids = "MP1,MP2"
        >>> )
        >>> response = layer.get_mount_point_props(request)
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        return self.stub.getMountPointsProperties(request._convert_to_grpc())

    @require_version(261)
    def update_mount_points(
        self, request: UpdateMountPointsRequest
    ) -> SherlockLayerService_pb2.UpdateMountPointsResponse:
        """Update mount point properties of a CCA from input parameters.

        Available Since: 2026R1

        Parameters
        ----------
        request: UpdateMountPointsRequest
            Contains all the information needed to update the properties for one or more
            mount points.

        Returns
        -------
        SherlockCommonService_pb2.UpdateMountPointsResponse
            A status code and message for the update mount points request.

        Examples
        --------
        >>> from ansys.sherlock.core.types.layer_types import UpdateMountPointsRequest,
        >>> MountPointProperties
        >>> from ansys.sherlock.core import launcher
        >>> sherlock, install_dir = launcher.launch_and_connect(transport_mode="wnua")
        >>> mount_point = MountPointProperties(
        >>>     id="MP1",
        >>>     type="Mount Pad",
        >>>     shape="Rectangular",
        >>>     units="mm",
        >>>     side="BOTTOM",
        >>>     height=1.0,
        >>>     material="GOLD",
        >>>     state="DISABLED",
        >>>     x=0.3,
        >>>     y=-0.4,
        >>>     length=1.0,
        >>>     width=0.2,
        >>>     diameter=0.0,
        >>>     nodes="",
        >>>     rotation=45,
        >>>     polygon="",
        >>>     boundary="Outline",
        >>>     constraints="X-axis translation|Z-axis translation",
        >>>     chassis_material="SILVER",
        >>> )
        >>> response = sherlock.layer.update_mount_points(UpdateMountPointsRequest(
        >>>     project="Tutorial Project",
        >>>     cca_name="Main Board",
        >>>     update_mount_points=[mount_point],
        >>> ))
        """
        if not self._is_connection_up():
            raise SherlockNoGrpcConnectionException()

        return self.stub.updateMountPoints(request._convert_to_grpc())



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2024 ANSYS, Inc. All rights reserved
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""pysherlock specific errors."""

from builtins import Exception
from typing import Optional

LOCALHOST = "127.0.0.1"
SHERLOCK_DEFAULT_PORT = 9090


class SherlockNoGrpcConnectionException(Exception):
    """Contains the error raised when the Sherlock gRPC channel has not been established."""

    def __init__(self):
        """Initialize error message."""
        self.message = "No connection to Sherlock gRPC server."

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockCannotUsePortError(Exception):
    """Contains the error raised when the specified gRPC port cannot be used."""

    def __init__(self, port: int, error: str):
        """Initialize error message."""
        self.port = port
        self.error = error

    def __str__(self):
        """Format error message."""
        return f"Specified gRPC port {self.port} cannot be used: {self.error}"


class SherlockConnectionError(Exception):
    """Contains the error raised when the Sherlock gRPC channel has not been established."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"gRPC connection error: {self.message}"


class SherlockDeleteProjectError(Exception):
    """Contains the error raised when a project cannot be deleted."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete project error: {self.message}"


class SherlockImportODBError(Exception):
    """Contains the error raised when an ODB archive cannot be imported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Import ODB error: {self.message}"


class SherlockImportIpc2581Error(Exception):
    """Contains the error raised when an IPC2581 archive cannot be imported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Import IPC2581 error: {self.message}"


class SherlockListCCAsError(Exception):
    """Contains the errors raised when a project's CCAs results cannot be listed."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.message is not None:
            return [f"List CCAs error: {self.message}"]
        elif self.error_array is not None:
            return [f"List CCAs error: {error}" for error in self.error_array]


class SherlockListConductorLayersError(Exception):
    """Contains the error raised when a project's conductor layers cannot be listed."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"List conductor layer error: {self.message}"


class SherlockListLaminateLayersError(Exception):
    """Contains the error raised when a project's laminate layers cannot be listed."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"List laminate layer error: {self.message}"


class SherlockGetStackupPropsError(Exception):
    """Contains the error raised when getting stackup properties."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Get stackup prop error: {self.message}"


class SherlockGetLayerCountError(Exception):
    """Contains the error raised when getting layer count."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Get layer count error: {self.message}"


class SherlockGetTotalConductorThicknessError(Exception):
    """Contains the error raised when getting total conductor thickness."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Get total conductor thickness error: {self.message}"


class SherlockAddStrainMapsError(Exception):
    """Contains the errors raised when strain maps cannot be added to the project."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add strain maps error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add strain maps error: {self.message}"]


class SherlockListStrainMapsError(Exception):
    """Contains the errors raised when strain maps for a project cannot be listed."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.message is not None:
            return [f"List strain maps error: {self.message}"]
        elif self.error_array is not None:
            return [f"List strain maps error: {error}" for error in self.error_array]


class SherlockGenerateProjectReportError(Exception):
    """Contains the error raised when project report cannot be generated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Generate project report error: {self.message}"


class SherlockCreateLifePhaseError(Exception):
    """Contains the errors raised when a life phase cannot be created."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.message is not None:
            return [f"Create life phase error: {self.message}"]

        elif self.error_array is not None:
            return [f"Create life phase error: {error}" for error in self.error_array]


class SherlockAddRandomVibeEventError(Exception):
    """Contains the errors raised when a random vibe event cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add random vibe event error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add random vibe event error: {self.message}"]


class SherlockAddRandomVibeProfilesError(Exception):
    """Contains the errors raised when random vibe profiles cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add random vibe profiles error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add random vibe profiles error: {self.message}"]


class SherlockAddThermalEventError(Exception):
    """Contains the errors raised when a thermal event cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add thermal event error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add thermal event error: {self.message}"]


class SherlockAddThermalProfilesError(Exception):
    """Creates the errors raised when thermal profiles cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add thermal profiles error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add thermal profiles error: {self.message}"]


class SherlockAddHarmonicEventError(Exception):
    """Contains the errors raised when a harmonic event cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add harmonic event error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add harmonic event error: {self.message}"]


class SherlockAddHarmonicVibeProfilesError(Exception):
    """Contains the errors raised when harmonic vibe profiles cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add harmonic vibe profiles error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add harmonic vibe profiles error: {self.message}"]


class SherlockAddShockEventError(Exception):
    """Contains the errors raised when a shock event cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add shock event error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add shock event error: {self.message}"]


class SherlockAddShockProfilesError(Exception):
    """Contains the errors raised when shock profiles cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Add shock profiles error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add shock profiles error: {self.message}"]


class SherlockLoadRandomVibeProfileError(Exception):
    """Contains the error raised when loading random vibe properties."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Load random vibe profile error: {self.message}"


class SherlockLoadHarmonicProfileError(Exception):
    """Contains the error raised when loading a harmonic profile."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Load harmonic profile error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Load harmonic profile error: {self.message}"]


class SherlockUpdateMountPointsByFileError(Exception):
    """Contains the errors raised when mount points cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update mount points by file error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update mount points by file error: {self.message}"]


class SherlockGenStackupError(Exception):
    """Contains the error raised when a stackup cannot be generated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Generate stackup error: {self.message}"


class SherlockUpdateConductorLayerError(Exception):
    """Contains the error raised when a conductor layer cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update conductor layer error: {self.message}"


class SherlockUpdateLaminateLayerError(Exception):
    """Contains the error raised when a laminate layer cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update laminate layer error: {self.message}"


class SherlockUpdatePartsListError(Exception):
    """Contains the errors raised when a parts list cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update parts list error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update parts list error: {self.message}"]


class SherlockUpdatePartsLocationsError(Exception):
    """Contains the errors raised when part locations cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update parts locations error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update parts locations error: {self.message}"]


class SherlockUpdatePartsLocationsByFileError(Exception):
    """Contains the errors raised when part locations cannot be updated by file results."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update parts locations by file error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update parts locations by file error: {self.message}"]


class SherlockImportPartsListError(Exception):
    """Contains the error raised when a parts list cannot be imported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Import parts list error: {self.message}"


class SherlockExportPartsListError(Exception):
    """Contains the error raised when a parts list cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export parts list error: {self.message}"


class SherlockEnableLeadModelingError(Exception):
    """Contains the error raised when lead modeling cannot be enabled."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Enable lead modeling error: {self.message}"


class SherlockLoadThermalProfileError(Exception):
    """Contains the error raised when loading thermal profile."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Load thermal profile error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Load thermal profile error: {self.message}"]


class SherlockRunAnalysisError(Exception):
    """Contains the error raised when an analysis cannot be run."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Run analysis error: {self.message}"


class SherlockRunStrainMapAnalysisError(Exception):
    """Contains the error raised when a strain map analysis cannot be run."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Run strain map analysis error: {self.message}"


class SherlockUpdateRandomVibePropsError(Exception):
    """Contains the error raised when properties for random vibe results cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update random vibe properties error: {self.message}"


class SherlockLoadShockProfileDatasetError(Exception):
    """Contains the error raised when loading shock profile dataset results in an error."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Load shock profile dataset error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Load shock profile dataset error: {self.message}"]


class SherlockUpdateNaturalFrequencyPropsError(Exception):
    """Contains the error raised when properties for natural frequency results cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update natural frequency properties error: {self.message}"


class SherlockCommonServiceError(Exception):
    """Contains the error raised when an API in the common service cannot be executed."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Sherlock common service error: {self.message}"


class SherlockModelServiceError(Exception):
    """Contains the error raised when an API in the model service cannot be executed."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Model service error: {self.message}"


class SherlockExportAEDBError(Exception):
    """Contains the error raised when an Electronics Desktop model cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export AEDB error: {self.message}"


class SherlockInvalidLoadDirectionError(Exception):
    """Contains the error raised when the load direction string is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidOrientationError(Exception):
    """Contains the error raised when an orientation string is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidRandomVibeProfileEntriesError(Exception):
    """Contains the error raised when a random vibe profile is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidThermalProfileEntriesError(Exception):
    """Contains the error raised when a thermal profile entry is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidHarmonicProfileEntriesError(Exception):
    """Contains the error raised when a harmonic profile entry is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidShockProfileEntriesError(Exception):
    """Contains the error raised when a shock profile entry is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidLayerIDError(Exception):
    """Contains the error raised when a layer ID is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidMaterialError(Exception):
    """Contains the error raised when a manufacturer/grade/material combination is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidConductorPercentError(Exception):
    """Contains the error raised when a conductor percent is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockInvalidGlassConstructionError(Exception):
    """Contains the error raised when the glass construction is invalid."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return self.message


class SherlockLoadShockProfilePulsesError(Exception):
    """Contains the error raised when loading shock profile pulses."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Load shock profile pulses error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Load shock profile pulses error: {self.message}"]


class SherlockUpdatePcbModelingPropsError(Exception):
    """Contains the error raised when updating pcb modeling properties results in an error."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Initialize error message."""
        return f"Update PCB Modeling Error: {self.message}"


class SherlockUpdateHarmonicVibePropsError(Exception):
    """Contains the error raised when properties for harmonic vibe analysis cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update harmonic vibe properties error: {self.message}"


class SherlockUpdateICTAnalysisPropsError(Exception):
    """Contains the error raised when properties for ICT analysis cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update ICT analysis properties error: {self.message}"


class SherlockUpdateMechanicalShockPropsError(Exception):
    """Contains the error raised when properties for mechanical shock analysis cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update mechanical shock properties error: {self.message}"


class SherlockUpdatePartListValidationAnalysisPropsError(Exception):
    """Contains the error raised when part list validation analysis properties cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update part list validation analysis properties error: {self.message}"


class SherlockUpdateSolderFatiguePropsError(Exception):
    """Contains the error raised when properties for solder fatigue analysis cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update solder fatigue properties error: {self.message}"


class SherlockAddProjectError(Exception):
    """Contains the error raised when Project cannot be added."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Add project error: {self.message}"


class SherlockAddCCAError(Exception):
    """Contains the error raised when CCA cannot be added."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Add CCA error: {self.message}"


class SherlockAddPottingRegionError(Exception):
    """Contains the error raised when a potting region cannot be added."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Add potting region error: {self.message}"


class SherlockUpdatePartModelingPropsError(Exception):
    """Contains the error raised when part modeling properties cannot be updated."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update part modeling props error: {self.message}"


class SherlockUpdatePartsFromAVLError(Exception):
    """Contains the error raised when parts list cannot be updated by AVL."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update part from AVL error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update part from AVL error: {self.message}"]


class SherlockListThermalMapsError(Exception):
    """Contains the errors raised when thermal map files for a project cannot be listed."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"List thermal maps error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"List thermal maps error: {self.message}"]


class SherlockUpdateThermalMapsError(Exception):
    """Contains the errors raised when thermal map files for a project cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Update thermal maps error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update thermal maps error: {self.message}"]


class SherlockAddThermalMapsError(Exception):
    """Contains the errors raised when thermal map files for a project cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Add thermal maps error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add thermal maps error: {self.message}"]


class SherlockImportProjectZipArchiveError(Exception):
    """Contains the error raised when a .zip project archive cannot be imported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Import zipped project archive error: {self.message}"


class SherlockImportProjectZipArchiveSingleModeError(Exception):
    """Contains the error raised when a .zip project archive cannot be imported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Import zipped project archive error: {self.message}"


class SherlockUpdatePartsListPropertiesError(Exception):
    """Contains the errors raised when a parts list properties cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Update parts list properties error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update parts list properties error: {self.message}"]


class SherlockExportNetListError(Exception):
    """Contains the error raised when a net list cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export net list error: {self.message}"


class SherlockExportProjectError(Exception):
    """Contains the error raised when a project cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export project error : {self.message}"


class SherlockDeleteAllMountPointsError(Exception):
    """Contains the error raised when the mount points cannot be deleted."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete mount points error: {self.message}"


class SherlockDeleteAllICTFixturesError(Exception):
    """Contains the error raised when the ICT fixtures cannot be deleted."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete ict fixtures error: {self.message}"


class SherlockDeleteAllTestPointsError(Exception):
    """Contains the error raised when the test points cannot be deleted."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete test points error: {self.message}"


class SherlockUpdateTestPointsByFileError(Exception):
    """Contains the errors raised when test points cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update test points by file error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update test points by file error: {self.message}"]


class SherlockUpdateTestFixturesByFileError(Exception):
    """Contains the errors raised when test fixtures cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Create list of error messages."""
        if self.error_array is not None:
            return [f"Update test fixtures by file error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update test fixtures by file error: {self.message}"]


class SherlockExportAllTestPointsError(Exception):
    """Contains the errors raised when test points cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export test points error: {self.message}"


class SherlockExportAllTestFixtures(Exception):
    """Contains the errors raised when test fixtures cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export test fixtures error: {self.message}"


class SherlockExportAllMountPoints(Exception):
    """Contains the errors raised when mount points cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export mount points error: {self.message}"


class SherlockCreateCCAFromModelingRegionError(Exception):
    """Contains the error raised when a CCA cannot be created from a modeling region."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Create CCA from modeling region error: {self.message}"


class SherlockExportFEAModelError(Exception):
    """Contains the error raised when a FEA model cannot be exported."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export FEA model error: {self.message}"


class SherlockAddModelingRegionError(Exception):
    """Contains the errors raised when modeling regions for a project cannot be added."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Add modeling region error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Add modeling region error: {self.message}"]


class SherlockUpdateModelingRegionError(Exception):
    """Contains the errors raised when modeling regions for a project cannot be updated."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Update modeling region error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Update modeling region error: {self.message}"]


class SherlockCopyModelingRegionError(Exception):
    """Contains the errors raised when modeling regions for a project cannot be copied."""

    def __init__(self, message: Optional[str] = None, error_array: Optional[list[str]] = None):
        """Initialize error message."""
        self.message = message
        self.error_array = error_array

    def str_itr(self):
        """Format error message."""
        if self.error_array is not None:
            return [f"Copy modeling region error: {error}" for error in self.error_array]

        elif self.message is not None:
            return [f"Copy modeling region error: {self.message}"]


class SherlockGetPartsListValidationAnalysisPropsError(Exception):
    """Contains the error raised when getting parts list validation properties."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Get parts list validation analysis properties error: {self.message}"


class SherlockDeleteModelingRegionError(Exception):
    """Contains the error raised when the modeling regions cannot be deleted."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete modeling region error: {self.message}"


class SherlockVersionError(Exception):
    """Contains the error raised when an incompatible Sherlock version is being used."""

    def __init__(self, message: str):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Sherlock Version Error: {self.message}"


class SherlockListLayersError(Exception):
    """Contains the errors raised when layers for a project CCA cannot be listed."""

    def __init__(self, message):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"List layers error: {self.message}"


class SherlockExportLayerImageError(Exception):
    """Contains the errors raised when 2D layer viewer image of CCA layers cannot be exported."""

    def __init__(self, message):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Export layer image error: {self.message}"


class SherlockSaveProfileError(Exception):
    """Contains the errors raised when a profile for an existing event cannot be saved."""

    def __init__(self, message):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Save profile error: {self.message}"


class SherlockDeleteError(Exception):
    """Contains the errors raised when an event or a phase cannot be deleted."""

    def __init__(self, message):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Delete error: {self.message}"


class SherlockUpdateLifePhaseError(Exception):
    """Contains the error raised when a given life phase cannot be updated."""

    def __init__(self, message):
        """Initialize error message."""
        self.message = message

    def __str__(self):
        """Format error message."""
        return f"Update life phase error: {self.message}"



# -*- coding: utf-8 -*-
#
# Copyright (C) 2021 - 2026 ANSYS, Inc. and/or its affiliates.
# © 2023 - 2025 ANSYS, Inc. All rights reserved.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

"""Module containing types for the Analysis Service."""

from enum import Enum
from typing import Optional

from pydantic import BaseModel, ConfigDict, ValidationInfo, field_validator

from ansys.sherlock.core.types.common_types import basic_str_validator

try:
    import SherlockAnalysisService_pb2
except ModuleNotFoundError:
    from ansys.api.sherlock.v0 import SherlockAnalysisService_pb2

analysis_service = SherlockAnalysisService_pb2


class ElementOrder:
    """Constants for Element Order."""

    LINEAR = analysis_service.ElementOrder.Linear
    "LINEAR"
    QUADRATIC = analysis_service.ElementOrder.Quadratic
    "QUADRATIC"
    SOLID_SHELL = analysis_service.ElementOrder.SolidShell
    "SOLID_SHELL"


class ModelSource:
    """Constants for Model Source."""

    GENERATED = analysis_service.ModelSource.GENERATED
    "GENERATED"
    STRAIN_MAP = analysis_service.ModelSource.STRAIN_MAP
    "STRAIN_MAP"


class RunAnalysisRequestAnalysisType:
    """Constants for type of analysis in the Run Analysis request."""

    __analysis_type = analysis_service.RunAnalysisRequest.Analysis.AnalysisType
    NATURAL_FREQ = __analysis_type.NaturalFreq
    "NATURAL_FREQ"
    HARMONIC_VIBE = __analysis_type.HarmonicVibe
    "HARMONIC_VIBE"
    ICT = __analysis_type.ICTAnalysis
    "ICT"
    MECHANICAL_SHOCK = __analysis_type.MechanicalShock
    "MECHANICAL_SHOCK"
    RANDOM_VIBE = __analysis_type.RandomVibe
    "RANDOM_VIBE"
    COMPONENT_FAILURE_MODE = __analysis_type.ComponentFailureMode
    "COMPONENT_FAILURE_MODE"
    DFMEA = __analysis_type.DFMEAModule
    "DFMEA"
    PTH_FATIQUE = __analysis_type.PTHFatigue
    "PTH_FATIQUE"
    PART_VALIDATION = __analysis_type.PartValidation
    "PART_VALIDATION"
    SEMICINDUCTOR_WEAROUT = __analysis_type.SemiconductorWearout
    "SEMICINDUCTOR_WEAROUT"
    SOLDER_JOINT_FATIGUE = __analysis_type.SolderJointFatigue
    "SOLDER_JOINT_FATIGUE"
    THERMAL_DERATING = __analysis_type.ThermalDerating
    "THERMAL_DERATING"
    THERMAL_MECH = __analysis_type.ThermalMech
    "THERMAL_MECH"


class RunStrainMapAnalysisRequestAnalysisType:
    """Constants for type of analysis in the Run Strain Map Analysis request."""

    __analysis_type = analysis_service.RunStrainMapAnalysisRequest.StrainMapAnalysis.AnalysisType
    HARMONIC_VIBE = __analysis_type.HarmonicVibe
    "HARMONIC_VIBE"

    MECHANICAL_SHOCK = __analysis_type.MechanicalShock
    "MECHANICAL_SHOCK"

    RANDOM_VIBE = __analysis_type.RandomVibe
    "RANDOM_VIBE"


class UpdatePcbModelingPropsRequestAnalysisType:
    """Constants for type of analysis in the Update PCB Modeling Properties Analysis request."""

    __analysis_type = analysis_service.UpdatePcbModelingPropsRequest.Analysis.AnalysisType
    HARMONIC_VIBE = __analysis_type.HarmonicVibe
    "HARMONIC_VIBE"
    ICT = __analysis_type.ICTAnalysis
    "ICT"
    MECHANICAL_SHOCK = __analysis_type.MechanicalShock
    "MECHANICAL_SHOCK"
    NATURAL_FREQUENCY = __analysis_type.NaturalFreq
    "NATURAL_FREQUENCY"
    RANDOM_VIBE = __analysis_type.RandomVibe
    "RANDOM_VIBE"
    THERMAL_MECH = __analysis_type.ThermalMech
    "THERMAL_MECH"


class UpdatePcbModelingPropsRequestPcbMaterialModel:
    """Constants for PCB Material Model in the Update PCB Modeling Properties Analysis request."""

    __material_model = analysis_service.UpdatePcbModelingPropsRequest.Analysis.PcbMaterialModel
    UNIFORM = __material_model.Uniform
    "UNIFORM"
    LAYERED = __material_model.Layered
    "LAYERED"
    UNIFORM_ELEMENTS = __material_model.UniformElements
    "UNIFORM_ELEMENTS"
    LAYERED_ELEMENTS = __material_model.LayeredElements
    "LAYERED_ELEMENTS"


class UpdatePcbModelingPropsRequestPcbModelType:
    """Constants for PCB Model Type in the Update PCB Modeling Properties Analysis request."""

    __model_type = analysis_service.UpdatePcbModelingPropsRequest.Analysis.PcbModelType
    BONDED = __model_type.Bonded
    "BONDED"


class ComponentFailureMechanism(BaseModel):
    """Contains the properties of a component failure mechanism update request."""

    cca_name: str
    """Name of the CCA."""
    default_part_temp_rise: float
    """Default part temperature rise."""
    default_part_temp_rise_units: str
    """Default part temperature rise units."""
    part_temp_rise_min_enabled: bool
    """Whether part temperature rise value is applied to the minimum temperature defined in the
    thermal cycle."""
    part_validation_enabled: bool
    """Whether part validation should be performed."""

    def _convert_to_grpc(
        self,
    ) -> analysis_service.UpdateComponentFailureMechanismPropsRequest.ComponentFailureMechanism:

        grpc_data = (
            analysis_service.UpdateComponentFailureMechanismPropsRequest.ComponentFailureMechanism()
        )

        grpc_data.ccaName = self.cca_name
        grpc_data.defaultPartTempRise = self.default_part_temp_rise
        grpc_data.defaultPartTempRiseUnits = self.default_part_temp_rise_units
        grpc_data.partTempRiseMinEnabled = self.part_temp_rise_min_enabled
        grpc_data.partValidationEnabled = self.part_validation_enabled
        return grpc_data

    @field_validator("cca_name")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdateComponentFailureMechanismPropsRequest(BaseModel):
    """Contains the properties of a component failure mechanism update per project."""

    project: str
    """Name of the Sherlock project."""
    component_failure_mechanism_properties_per_cca: list[ComponentFailureMechanism]
    """List of potting region data to update."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> analysis_service.UpdateComponentFailureMechanismPropsRequest:
        request = analysis_service.UpdateComponentFailureMechanismPropsRequest()
        request.project = self.project
        for properties in self.component_failure_mechanism_properties_per_cca:
            request.componentFailureMechanismProperties.append(properties._convert_to_grpc())
        return request


class SemiconductorWearoutAnalysis(BaseModel):
    """Contains the properties of a semiconductor wearout analysis update request."""

    cca_name: str
    """Name of the CCA."""
    max_feature_size: float
    """Maximum feature size."""
    max_feature_size_units: str
    """Units of the maximum feature size."""
    part_temp_rise: float
    """Part temperature rise."""
    part_temp_rise_units: str
    """Units of the part temperature rise."""
    part_temp_rise_min_enabled: bool
    """Whether part temperature rise value is applied to the minimum temperature defined in the
    thermal cycle."""
    part_validation_enabled: bool
    """Whether part validation should be performed."""

    def _convert_to_grpc(
        self,
    ) -> (
        analysis_service.UpdateSemiconductorWearoutAnalysisPropsRequest.SemiconductorWearoutAnalysis
    ):

        request_class = analysis_service.UpdateSemiconductorWearoutAnalysisPropsRequest
        semiconductor_wearout_analysis = request_class.SemiconductorWearoutAnalysis
        grpc_data = semiconductor_wearout_analysis()

        grpc_data.ccaName = self.cca_name
        grpc_data.maxFeatureSize = self.max_feature_size
        grpc_data.maxFeatureSizeUnits = self.max_feature_size_units
        grpc_data.partTempRise = self.part_temp_rise
        grpc_data.partTempRiseUnits = self.part_temp_rise_units
        grpc_data.partTempRiseMinEnabled = self.part_temp_rise_min_enabled
        grpc_data.partValidationEnabled = self.part_validation_enabled
        return grpc_data

    @field_validator("cca_name", "max_feature_size_units", "part_temp_rise_units")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdateSemiconductorWearoutAnalysisPropsRequest(BaseModel):
    """Contains the properties of a semiconductor wearout analysis update per project."""

    project: str
    """Name of the Sherlock project."""
    semiconductor_wearout_analysis_properties: list[SemiconductorWearoutAnalysis]
    """List of semiconductor wearout analysis properties to update."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> analysis_service.UpdateSemiconductorWearoutAnalysisPropsRequest:
        request = analysis_service.UpdateSemiconductorWearoutAnalysisPropsRequest()
        request.project = self.project
        for properties in self.semiconductor_wearout_analysis_properties:
            request.semiconductorWearoutAnalysisProperties.append(properties._convert_to_grpc())
        return request


class UpdatePTHFatiguePropsRequestAnalysisType(Enum):
    """Constants for qualification choices in the Update PTH Fatigue Properties request."""

    __qualification = analysis_service.UpdatePTHFatiguePropsRequest.PTHFatigueAnalysis.Qualification
    NONE = __qualification.NONE
    "NONE"
    PER_LOT = __qualification.PER_LOT
    "PER_LOT"
    PRODUCT = __qualification.PRODUCT
    "PRODUCT"
    SUPPLIER = __qualification.SUPPLIER
    "SUPPLIER"


class PTHFatiguePropsAnalysis(BaseModel):
    """Contains the properties of a PTH fatigue analysis update request."""

    model_config = ConfigDict(arbitrary_types_allowed=True)

    cca_name: str
    """Name of the CCA."""
    qualification: Optional[UpdatePTHFatiguePropsRequestAnalysisType] = None
    """Qualification choice for IST/HATS."""  # noqa: E501
    pth_quality_factor: Optional[str] = None
    """Quality factor for PTH."""
    pth_wall_thickness: Optional[float] = None
    """Wall thickness of PTH."""
    pth_wall_thickness_units: Optional[str] = None
    """Units for PTH wall thickness."""
    min_hole_size: Optional[float] = None
    """Minimum hole size."""
    min_hole_size_units: Optional[str] = None
    """Units for minimum hole size."""
    max_hole_size: Optional[float] = None
    """Maximum hole size."""
    max_hole_size_units: Optional[str] = None
    """Units for maximum hole size."""

    def _convert_to_grpc(
        self,
    ) -> analysis_service.UpdatePTHFatiguePropsRequest.PTHFatigueAnalysis:
        grpc_data = analysis_service.UpdatePTHFatiguePropsRequest.PTHFatigueAnalysis()

        grpc_data.ccaName = self.cca_name
        if self.qualification is not None:
            grpc_data.qualification = self.qualification.value
        if self.pth_quality_factor is not None:
            grpc_data.pthQualityFactor = self.pth_quality_factor
        if self.pth_wall_thickness is not None:
            grpc_data.pthWallThickness = self.pth_wall_thickness
        if self.pth_wall_thickness_units is not None:
            grpc_data.pthWallThicknessUnits = self.pth_wall_thickness_units
        if self.min_hole_size is not None:
            grpc_data.minHoleSize = self.min_hole_size
        if self.min_hole_size_units is not None:
            grpc_data.minHoleSizeUnits = self.min_hole_size_units
        if self.max_hole_size is not None:
            grpc_data.maxHoleSize = self.max_hole_size
        if self.max_hole_size_units is not None:
            grpc_data.maxHoleSizeUnits = self.max_hole_size_units

        return grpc_data

    @field_validator(
        "cca_name",
        "pth_quality_factor",
        "pth_wall_thickness_units",
        "min_hole_size_units",
        "max_hole_size_units",
    )
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)


class UpdatePTHFatiguePropsRequest(BaseModel):
    """Contains the properties of a PTH fatigue analysis update per project."""

    project: str
    """Name of the Sherlock project."""
    pth_fatigue_analysis_properties: list[PTHFatiguePropsAnalysis]
    """List of PTH fatigue analysis properties to update."""

    @field_validator("project")
    @classmethod
    def str_validation(cls, value: str, info: ValidationInfo):
        """Validate string fields listed."""
        return basic_str_validator(value, info.field_name)

    def _convert_to_grpc(
        self,
    ) -> analysis_service.UpdatePTHFatiguePropsRequest:
        request = analysis_service.UpdatePTHFatiguePropsRequest()
        request.project = self.project
        for properties in self.pth_fatigue_analysis_properties:
            request.pthFatigueAnalysisProperties.append(properties._convert_to_grpc())
        return request
